package AminoAcid;

import Alignments.Alignment;
import Alignments.Ambiguous;
import Alignments.PhylipAlignment;
import Exceptions.GeneralException;
import Likelihood.StandardCalculator;
import Likelihood.StandardLikelihood;
import Models.Model;
import Models.RateCategory;
import Optimizers.GoldenSection;
import Optimizers.Optimizer;
import Parameters.Parameter;
import Parameters.Parameters;
import Trees.Tree;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

public class AminoAcid
{

    public static void main(String[] args) throws GeneralException, IOException
    {
        /*
         * Implements the LG4X model using GeLL
         */
        
        /*
        * Setting the parameter values to be the same as those from the phyml-4X
        * output (included in the data directory), as we do here, results in a near identical
        * likelihood.  If we optimize the parameters we end up with a better
        * likelihood and slightly different parameter estimates.  This is
        * probably due to the lazy optimization used in phyml.
        */
        
        //Create a hash map for the different rate categories ... i.e. X1 through X4
        HashMap<RateCategory,String> freq = new HashMap<>();
        for (int i = 1; i <= 4; i++)
        {
            //Read in each rate file and create the category
            RateCategory r = getRateCategory(new File("src\\AminoAcid\\data\\X" + i + ".dat")).multiplyBy("r" + i);
            //Set the name of the category - useful for output
            r.setName("X" + i);
            //Associate a frequency with the rate category.  For this model
            //we estimate the frequencies so the frequency is a different
            //parameter for each category.  In this case the parameters are
            //called f1..f4
            freq.put(r,"f" + i);
        }
        
        //Create the model
        Model m = new Model(freq);
        
        //Read in the tree
        Tree t = Tree.fromFile(new File("C:\\Temp\\bpptest\\AA\\outtree.dat"));
        
        //Create a set of all amino acid characters
        String all = "A R N D C Q E G H I L K M F P S T W Y V";
        String[] split = all.split(" ");
        HashSet<String> set = new HashSet<>();
        for (String s: split)
        {
            set.add(s);
        }
        //Create a map from an ambiguous character (in this case "-") to the set
        //all possible characters it can represent (in this case every amino
        //acid).  Then create the ambiguous object
        Map<String,Set<String>> am = new HashMap<>();
        am.put("-",set);
        Ambiguous ambig = new Ambiguous(am);
        
        //Read in the alignment including passing details of ambiguous characters
        Alignment a = PhylipAlignment.fromFile(new File("src\\AminoAcid\\data\\alignment.dat"), ambig);
        
        //Create the calculator
        StandardCalculator c = new StandardCalculator(m,a,t);
        
        //Set parameters to the values from Phyml-4X   
        Parameters p = t.getParameters();
        p.addParameter(Parameter.newFixedParameter("f1", 0.440066));
        p.addParameter(Parameter.newFixedParameter("f2", 0.335967));
        p.addParameter(Parameter.newFixedParameter("f3", 0.190398));
        p.addParameter(Parameter.newFixedParameter("f4", 0.033568));

        p.addParameter(Parameter.newFixedParameter("r1", 0.147510));
        p.addParameter(Parameter.newFixedParameter("r2", 0.939876));
        p.addParameter(Parameter.newFixedParameter("r3", 0.978432));
        p.addParameter(Parameter.newFixedParameter("r4", 12.899838));

        StandardLikelihood l = c.calculate(p);
         
        //Print out the likelihood
        System.out.println(l.getLikelihood());
    }
 
    private static RateCategory getRateCategory(File f) throws GeneralException, IOException
    {
        //Reads in the defintion of one of the rate categories from file
        BufferedReader in = new BufferedReader(new FileReader(f));

        String[][] m = new String[20][20];
        
        //The diagonals aren't used in the calculations so set them to dummy values
        for (int i = 0; i < 20; i++)
        {
            m[i][i] = "";
        }
        
        //Read in each line and assign values as appropiate - to both the
        //position in the file and the "opposite" position in the matrix
        for (int i = 1; i < 20; i++)
        {
            String line = in.readLine();
            
            String[] parts = line.split("\\s+");
            
            for (int j = 0; j < parts.length; j++)
            {
                m[i][j] = parts[j];
                m[j][i] = parts[j];
            }
        }
        
        in.readLine();
        
        //Read in the frequencies
        String[] freq = in.readLine().split("\\s+");
        
        //Multiply each entry in the array by the appropiate frequency
        for (int i = 0; i < 20; i++)
        {
            for (int j = 0; j < 20; j++)
            {
                if (i != j)
                {
                    m[i][j] = m[i][j] + " * " + freq[j];
                }
            }
        }
        
        //Create a map from character to position in the array
        String order = "A R N D C Q E G H I L K M F P S T W Y V";
        String[] o = order.split(" ");
        Map<String,Integer> map = new HashMap<>();
        
        for (int i = 0; i < o.length; i++)
        {
            map.put(o[i], i);
        }
        
        //Create the rate category
        return new RateCategory(m,freq,map);
    }
}
