package Codon;

import Alignments.Alignment;
import Alignments.PhylipAlignment;
import Alignments.Site;
import Exceptions.GeneralException;
import Likelihood.Likelihood;
import Likelihood.StandardCalculator;
import Models.Model;
import Models.RateCategory;
import Optimizers.GoldenSection;
import Optimizers.Optimizer;
import Parameters.Parameter;
import Parameters.Parameters;
import Trees.Tree;
import java.io.File;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

public class Codon
{
    public static void main(String[] args) throws GeneralException
    {
        //Load the alignment as a DNA alignment
        Alignment oa = PhylipAlignment.fromFile(new File("src\\Codon\\data\\alignment.dat"));
        //Convert it to a codon alignment
        Alignment a = convert(oa);
        
        //Get the frequencies
        Map<String,Double> freq = getFreq(a);
        
        //Get the model
        Model m = model(freq);
        //Create the tree
        Tree t = Tree.fromNewickString("((Hsa_Human: 0.025240, Hla_gibbon: 0.041269): 0.077423, ((Cgu/Can_colobus: 0.044409, Pne_langur: 0.053758): 0.085025, Mmu_rhesus: 0.019175): 0.046023, (Ssc_squirrelM: 0.042962, Cja_marmoset: 0.024797): 0.136928);");
        
        Parameters p = t.getParametersForEstimation();
        p.addParameter(Parameter.newEstimatedBoundedParameter("o1", 0.0, 1.0));
        p.addParameter(Parameter.newFixedParameter("o2", 1.0));
        p.addParameter(Parameter.newEstimatedBoundedParameter("o3", 1.0, Double.MAX_VALUE));
        p.addParameter(Parameter.newEstimatedPositiveParameter("k"));
        p.addParameter(Parameter.newFixedParameter("f1", 1.0));
        p.addParameter(Parameter.newEstimatedPositiveParameter("f2"));
        p.addParameter(Parameter.newEstimatedPositiveParameter("f3"));
        
        
        //Create the optimizer - this prints out output every time it optimizes
        //a parameter
        Optimizer o = new GoldenSection(GoldenSection.ProgressLevel.PARAMETER);
        
        //Create the calculator
        StandardCalculator c = new StandardCalculator(m,a,t);
        
        //Maximise the likelihood
        Likelihood l = o.maximise(c, p);
        
        //Print out the estimated parameters and likelihoods
        System.out.println("\n\n");
        System.out.println(l.getLikelihood());
        System.out.println();
        System.out.println(l.getParameters());
    }
    
    private static Alignment convert(Alignment oa) throws GeneralException
    {
        //Create a list to house the codon "sites"
        List<Site> sites = new ArrayList<>(oa.getLength() / 3);
        //For every triplet of DNA sites in the alignment
        for (int i = 0; i < oa.getLength() / 3; i++)
        {
            //Get the three DNA sites
            Site p1 = oa.getSite(i*3);
            Site p2 = oa.getSite(i*3+1);
            Site p3 = oa.getSite(i*3+2);
            
            //Create the hashmap to hold the charcaters for the new codon "site"
            LinkedHashMap<String,String> site = new LinkedHashMap<>();
            //For each taxa
            for (String t: oa.getTaxa())
            {
                //Add the new codon "site" - which is simply the three DNA
                //characters combined
                site.put(t,p1.getRawCharacter(t) + p2.getRawCharacter(t) +
                        p3.getRawCharacter(t));
            }
            //Create and add the new codon "site" to the list of sites
            sites.add(new Site(site));
        }
        //Create and return the alignment
        return new Alignment(sites);
    }
    
    private static Model model(Map<String,Double> ofreq) throws GeneralException
    {
        String[] nuc = {"A","C","T","G"};
        HashMap<String,Integer> map = new HashMap<>();
        String[] rmap = new String[64];
        
        String[] freq = new String[64];
       
        int c = 0;
        for (String n1: nuc)
        {
            for (String n2: nuc)
            {
                for (String n3: nuc)
                {
                    map.put(n1 + n2 + n3,c);
                    rmap[c] = n1 + n2 + n3;
                    
                    if (cmap.get(n1 + n2 + n3).equals("*"))
                    {
                        freq[c] = "0.0";
                    }
                    else
                    {
                        freq[c] = Double.toString(ofreq.get(n1 + n2 + n3));
                    }
                    c++;
                }
            }
        }
        
        HashMap<RateCategory,String> rcfreq = new HashMap<>();
        for (int i = 1; i <= 3; i++)
        {
            rcfreq.put(rc(freq,"o" + i, map, rmap), "f" + i);
        }
        
        return new Model(rcfreq);
    }
    
    private static Map<String,Double> getFreq(Alignment a) throws GeneralException
    {
        Map<Character,Integer> nmap = new HashMap<>();
        nmap.put('A',0); nmap.put('C',1); nmap.put('T',2); nmap.put('G',3);
        
        int f[][] = new int[3][4];

        for (Site s: a)
        {
            for (String t: a.getTaxa())
            {
                String codon = s.getRawCharacter(t);
                char[] nuc = codon.toCharArray();

                for (int i = 0; i < 3; i++)
                {
                    int n = nmap.get(nuc[i]);
                    f[i][n]++;
                }
            }
        }
        
        Map<String,Double> rfreq = new HashMap<>();
        double t = a.getLength() * a.getNumber();
        
        for (Entry<Character,Integer> e1: nmap.entrySet())
        {
            for (Entry<Character,Integer> e2: nmap.entrySet())
            {
                for (Entry<Character,Integer> e3: nmap.entrySet())
                {
                    String codon = e1.getKey().toString() + 
                            e2.getKey().toString() + e3.getKey().toString();
                    if (cmap.get(codon).equals("*"))
                    {
                        rfreq.put(codon, 0.0);
                    }
                    else
                    {
                        double f1 = (double) f[0][e1.getValue()] / t;
                        double f2 = (double) f[1][e2.getValue()] / t;
                        double f3 = (double) f[2][e3.getValue()] / t;
                        rfreq.put(codon,f1*f2*f3);
                    }
                }
            }
        }
        
        double ct = 0.0;
        for (double cf: rfreq.values())
        {
            ct += cf;
        }

        Map<String,Double> freq = new HashMap<>();
        for (Entry<String,Double> e: rfreq.entrySet())
        {
            freq.put(e.getKey(),e.getValue() / ct);
        }
        
        return freq;
    }
    
    private static RateCategory rc(String[] freq, String omega, Map<String,Integer> map, String[] rmap) throws GeneralException
    {
        String[][] array = new String[64][64];
        for (int i = 0; i < 64; i++)
        {
            for (int j = 0; j < 64; j++)
            {
                array[i][j] = makeRate(rmap[i],rmap[j],omega,freq[j]);
            }
        }
        
        RateCategory rc = new RateCategory(array,freq,map);
        rc.setName(omega);
        return rc;
    }
    
    private static String makeRate(String c1, String c2, String omega, String f)
    {
        switch (changeType(c1,c2))
        {
            case NO_CHANGE:
                return " ";
            case SYNONYMOUS_TRANSVERSION:
                return f;
            case SYNONYMOUS_TRANSITION:
                return f + " * k";
            case NONSYNONYMOUS_TRANSVERSION:
                return f + " * " + omega;
            case NONSYNONYMOUS_TRANSITION:
                return f + " * " + omega + " * k";
            case MORE_THAN_ONE:
            case INVALID:
            default:
                return "0"; 
        }
    }
    
    private static Map<String,String> codonMap()
    {
        Map<String,String> map = new HashMap<>();
        map.put("TTT","F");
        map.put("TTC","F");
        map.put("TTA","L");
        map.put("TTG","L");
        map.put("TCT","S");
        map.put("TCC","S");
        map.put("TCA","S");
        map.put("TCG","S");
        map.put("TAT","Y");
        map.put("TAC","Y");
        map.put("TAA","*");
        map.put("TAG","*");
        map.put("TGT","C");
        map.put("TGC","C");
        map.put("TGA","*");
        map.put("TGG","W");
        map.put("CTT","L");
        map.put("CTC","L");
        map.put("CTA","L");
        map.put("CTG","L");
        map.put("CCT","P");
        map.put("CCC","P");
        map.put("CCA","P");
        map.put("CCG","P");
        map.put("CAT","H");
        map.put("CAC","H");
        map.put("CAA","Q");
        map.put("CAG","Q");
        map.put("CGT","R");
        map.put("CGC","R");
        map.put("CGA","R");
        map.put("CGG","R");
        map.put("ATT","I");
        map.put("ATC","I");
        map.put("ATA","I");
        map.put("ATG","M");
        map.put("ACT","T");
        map.put("ACC","T");
        map.put("ACA","T");
        map.put("ACG","T");
        map.put("AAT","N");
        map.put("AAC","N");
        map.put("AAA","K");
        map.put("AAG","K");
        map.put("AGT","S");
        map.put("AGC","S");
        map.put("AGA","R");
        map.put("AGG","R");
        map.put("GTT","V");
        map.put("GTC","V");
        map.put("GTA","V");
        map.put("GTG","V");
        map.put("GCT","A");
        map.put("GCC","A");
        map.put("GCA","A");
        map.put("GCG","A");
        map.put("GAT","D");
        map.put("GAC","D");
        map.put("GAA","E");
        map.put("GAG","E");
        map.put("GGT","G");
        map.put("GGC","G");
        map.put("GGA","G");
        map.put("GGG","G");
        
        return map;
    }
    
    private static ChangeType changeType(String c1, String c2)
    {
        if (cmap.get(c1).equals("*") || cmap.get(c2).equals("*"))
        {
            return ChangeType.INVALID;
        }
        char[] ca1 = c1.toCharArray();
        char[] ca2 = c2.toCharArray();
        
        char change1 = ' ';
        char change2 = ' ';
        
        int c = 0;
        
        for (int i = 0; i < 3; i++)
        {
            if (ca1[i] != ca2[i])
            {
                c++;
                change1 = ca1[i];
                change2 = ca2[i];
            }
        }
        
        if (c == 0)
        {
            return ChangeType.NO_CHANGE;
        }
        if (c > 1)
        {
            return ChangeType.MORE_THAN_ONE;
        }
        if ( (change1 == 'A' && change2 == 'G') ||
             (change1 == 'G' && change2 == 'A') ||
             (change1 == 'C' && change2 == 'T') ||
             (change1 == 'T' && change2 == 'C') )
        {
            // TRANSITION
            if (cmap.get(c1).equals(cmap.get(c2)))
            {
                return ChangeType.SYNONYMOUS_TRANSITION;
            }
            else
            {
                return ChangeType.NONSYNONYMOUS_TRANSITION;
            }
        }
        else
        {
            // TRANSVERSION
            if (cmap.get(c1).equals(cmap.get(c2)))
            {
                return ChangeType.SYNONYMOUS_TRANSVERSION;
            }
            else
            {
                return ChangeType.NONSYNONYMOUS_TRANSVERSION;
            }            
        }
    }
    
    private enum ChangeType
    {
        NO_CHANGE,
        MORE_THAN_ONE,
        SYNONYMOUS_TRANSITION,
        SYNONYMOUS_TRANSVERSION,
        NONSYNONYMOUS_TRANSITION,
        NONSYNONYMOUS_TRANSVERSION,
        INVALID
    }
    
    private static final Map<String,String> cmap = codonMap();
}
