package WAG;

import Alignments.Alignment;
import Alignments.Ambiguous;
import Alignments.PhylipAlignment;
import Exceptions.GeneralException;
import Likelihood.StandardCalculator;
import Likelihood.StandardLikelihood;
import Models.Model;
import Models.RateCategory;
import Optimizers.GoldenSection;
import Optimizers.Optimizer;
import Parameters.Parameters;
import Trees.Tree;
import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

public class WAG
{

    public static void main(String[] args) throws GeneralException, IOException
    {
        /*
         * Implements the WAG model
         */  

        //Read in rate file and create the category
        RateCategory r = getRateCategory(new File("src\\WAG\\data\\wag.dat"));
        
        //Create the model
        Model m = new Model(r);
        
        //Read in the tree
        Tree t = Tree.fromFile(new File("src\\WAG\\data\\tree.dat"));
        
        //Create a set of all amino acid characters
        String all = "A R N D C Q E G H I L K M F P S T W Y V";
        String[] split = all.split(" ");
        HashSet<String> set = new HashSet<>();
        for (String s: split)
        {
            set.add(s);
        }
        //Create a map from an ambiguous character (in this case "?") to the set
        //all possible characters it can represent (in this case every amino
        //acid).  Then create the ambiguous object
        Map<String,Set<String>> am = new HashMap<>();
        am.put("?",set);
        Ambiguous ambig = new Ambiguous(am);
        
        //Read in the alignment including passing details of ambiguous characters
        Alignment a = PhylipAlignment.fromFile(new File("src\\WAG\\data\\alignment.dat"), ambig);
        
        //Create the calculator
        StandardCalculator c = new StandardCalculator(m,a,t);

 
        Parameters p = t.getParametersForEstimation();

        //Create the optimizer and optimize
        Optimizer o = new GoldenSection(GoldenSection.ProgressLevel.PARAMETER);
        StandardLikelihood l = o.maximise(c, p);

        //Print out the optimized parameters
        System.out.println(l.getParameters());
        System.out.println();
        
        //Print out the likelihood
        System.out.println(l.getLikelihood());
    }
 
    private static RateCategory getRateCategory(File f) throws GeneralException, IOException
    {
        //Reads in the defintion of one of the rate categories from file
        BufferedReader in = new BufferedReader(new FileReader(f));

        String[][] m = new String[20][20];
        
        //The diagonals aren't used in the calculations so set them to dummy values
        for (int i = 0; i < 20; i++)
        {
            m[i][i] = "";
        }
        
        //Read in each line and assign values as appropiate - to both the
        //position in the file and the "opposite" position in the matrix
        for (int i = 1; i < 20; i++)
        {
            String line = in.readLine();
            
            String[] parts = line.split("\\s+");
            
            for (int j = 0; j < parts.length; j++)
            {
                m[i][j] = parts[j];
                m[j][i] = parts[j];
            }
        }
        
        in.readLine();
        
        //Read in the frequencies
        String[] freq = in.readLine().split("\\s+");
        
        //Multiply each entry in the array by the appropiate frequency
        for (int i = 0; i < 20; i++)
        {
            for (int j = 0; j < 20; j++)
            {
                if (i != j)
                {
                    m[i][j] = m[i][j] + " * " + freq[j];
                }
            }
        }
        
        //Create a map from character to position in the array
        String order = "A R N D C Q E G H I L K M F P S T W Y V";
        String[] o = order.split(" ");
        Map<String,Integer> map = new HashMap<>();
        
        for (int i = 0; i < o.length; i++)
        {
            map.put(o[i], i);
        }
        
        //Create the rate category
        return new RateCategory(m,freq,map);
    }
}
