package DNA;

import Alignments.Alignment;
import Alignments.PhylipAlignment;
import Ancestors.AncestralJoint;
import Exceptions.GeneralException;
import Likelihood.StandardCalculator;
import Likelihood.StandardLikelihood;
import Models.Model;
import Models.RateCategory;
import Optimizers.GoldenSection;
import Optimizers.Optimizer;
import Parameters.Parameter;
import Parameters.Parameters;
import Simulations.Simulate;
import Trees.Tree;
import java.io.File;
import java.util.HashMap;
import java.util.Map;

public class DNA
{
    public static void main(String[] args) throws GeneralException
    {
        //Create a tree object.  Here we are going to estimate branch lengths so
        //no point including them in the tree
        Tree t = Tree.fromNewickString("(((Human, Chimpanzee)A, Gorilla)B, Orangutan, Gibbon)C;");
        //Load an alignment
        Alignment a = PhylipAlignment.fromFile(new File("src\\DNA\\data\\brown.nuc"));
        //Get the parameters from the tree.  This step is recommended although due
        //to it's weirdness will be done automatically if missed out.
        Parameters p = t.getParametersForEstimation();
        
        //Create the map that contains the rate categories and their frequency
        Map<RateCategory,String> rcs = new HashMap<>();
        //For each of the four rate categories we wish to create
        for (int i = 1; i <=4 ; i++)
        {
            //Create a string to represent the gamma function - g is the parameter,
            //i is the ith rate category, and 4 is the total number of categorys
            String g = "g[g," + i + ",4]*";
            
            //Create the rate matrix, adding the gamma value to the product
            String[][] ma = new String[4][4];
            ma[0][0] = "-"; ma[0][1] = g + "a*pC"; ma[0][2] = g + "b*pA"; ma[0][3] = g + "c*pG";
            ma[1][0] = g + "a*pT"; ma[1][1] = "-"; ma[1][2] = g + "d*pA"; ma[1][3] = g + "e*pG";
            ma[2][0] = g + "b*pT"; ma[2][1] = g + "d*pC"; ma[2][2] = "-"; ma[2][3] = g + "pG";
            ma[3][0] = g + "c*pT"; ma[3][1] = g + "e*pC"; ma[3][2] = g + "pA"; ma[3][3] = "-";

            //We're estimating character frequencies so create the array of
            //character frequencies with one parameter for each frequency
            String[] freq = {"pT", "pC", "pA", "pG"};

            //Create the map from character to position in the matrix
            HashMap<String,Integer> map = new HashMap<>();
            map.put("T",0);
            map.put("C",1);
            map.put("A",2);
            map.put("G",3);
            
            //Create the rate catgory and give it a name
            RateCategory rc = new RateCategory(ma,freq,map);
            rc.setName("Gamma Rate " + i);
            
            //Add it to the rate category map with the appropiate frequency
            rcs.put(rc,"0.25");
        }
        //Create the model
        Model m = new Model(rcs);
        
        //Create the parameters of the matrix and add them to our list of parameters
        p.addParameter(Parameter.newEstimatedPositiveParameter("a"));
        p.addParameter(Parameter.newEstimatedPositiveParameter("b"));
        p.addParameter(Parameter.newEstimatedPositiveParameter("c"));
        p.addParameter(Parameter.newEstimatedPositiveParameter("d"));
        p.addParameter(Parameter.newEstimatedPositiveParameter("e"));

        //And the frequency.  One frequency is set to one and the other frequencies
        //are then scaled to this.
        p.addParameter(Parameter.newFixedParameter("pT",1.0));
        p.addParameter(Parameter.newEstimatedPositiveParameter("pC"));
        p.addParameter(Parameter.newEstimatedPositiveParameter("pA"));
        p.addParameter(Parameter.newEstimatedPositiveParameter("pG"));
        
        //Add the gamma parameter
        p.addParameter(Parameter.newEstimatedPositiveParameter("g"));
        
        //Create a calculator which is used to calculate likelihoods
        StandardCalculator c = new StandardCalculator(m,a,t);
        //Create an optimizer to do the the optimization
        Optimizer o = new GoldenSection();
        
        //Actually do the optimization and get the result.  The parameters passed
        //will be updated to their estimated values.  The result will contain the
        //lieklihood and intermediate results in calculating it.
        StandardLikelihood l = o.maximise(c, p);
        
        //Update paameters to the optimized values
        p = l.getParameters();
        
        //Print out the likelihood and the estimated parameters
        System.out.println("Likelihood: " + l.getLikelihood());
        System.out.println();
        System.out.println("Parameters:");
        System.out.println(p);
        
        //Create an ancestral reconstruction object
        AncestralJoint aj = AncestralJoint.newInstance(m, a, t);
        
        //And do the reconstruction using the parameters we've just edstimated
        Alignment anc = aj.calculate(p);
        
        //Write out the reconstructed alignment
        PhylipAlignment.writeFile(anc, new File("src\\DNA\\data\\ancestor.dat"));
        
        //Create a simulation object, using the parameters we've just estimated
        Simulate s = new Simulate(m,t,p);
        
        //Create a simulated alignment
        Alignment sim = s.getAlignment(500);
        
        //Write out the simulated alignment
        PhylipAlignment.writeFile(sim, new File("src\\DNA\\data\\simulated.dat"));
    }
    
}
